#ifndef DPP_H
#define DPP_H

#define PUB_AF_CATEGORY		0x04
#define DPP_PUB_AF_ACTION		0x09
#define DPP_PUB_AF_ACTION_OUI_TYPE	0x1A
#define DPP_PUB_AF_GAS_INITIAL_REQ	0x0A
#define DPP_PUB_AF_GAS_INITIAL_RESP	0x0B

#define FRAME_IS_DPP_PUB_AF(frame) \
		((*(frame) == DPP_PUB_AF_ACTION) && \
		(memcmp((frame + 1), "\x50\x6F\x9A", 3) == 0) && \
		((*(frame + 4)) == DPP_PUB_AF_ACTION_OUI_TYPE))

#define FRAME_IS_DPP_GAS_FRAME(frame) \
		((*(frame) == DPP_PUB_AF_GAS_INITIAL_REQ) || \
		(*(frame) == DPP_PUB_AF_GAS_INITIAL_RESP))


#define FRAME_DST_ENROLLEE(type) \
		((type == DPP_PA_AUTHENTICATION_REQ) || \
		 (type == DPP_PA_AUTHENTICATION_CONF))

#define DPP_URI_FILE "/etc/multiap/dpp_uris.json"

enum dpp_public_action_frame_type {
	DPP_PA_AUTHENTICATION_REQ = 0,
	DPP_PA_AUTHENTICATION_RESP = 1,
	DPP_PA_AUTHENTICATION_CONF = 2,
	DPP_PA_PEER_DISCOVERY_REQ = 5,
	DPP_PA_PEER_DISCOVERY_RESP = 6,
	DPP_PA_PKEX_V1_EXCHANGE_REQ = 7,
	DPP_PA_PKEX_EXCHANGE_RESP = 8,
	DPP_PA_PKEX_COMMIT_REVEAL_REQ = 9,
	DPP_PA_PKEX_COMMIT_REVEAL_RESP = 10,
	DPP_PA_CONFIGURATION_RESULT = 11,
	DPP_PA_CONNECTION_STATUS_RESULT = 12,
	DPP_PA_PRESENCE_ANNOUNCEMENT = 13,
	DPP_PA_RECONFIG_ANNOUNCEMENT = 14,
	DPP_PA_RECONFIG_AUTH_REQ = 15,
	DPP_PA_RECONFIG_AUTH_RESP = 16,
	DPP_PA_RECONFIG_AUTH_CONF = 17,
	DPP_PA_PKEX_EXCHANGE_REQ = 18,
};

enum dpp_attribute_id {
	DPP_ATTR_STATUS = 0x1000,
	DPP_ATTR_I_BOOTSTRAP_KEY_HASH = 0x1001,
	DPP_ATTR_R_BOOTSTRAP_KEY_HASH = 0x1002,
	DPP_ATTR_I_PROTOCOL_KEY = 0x1003,
	DPP_ATTR_WRAPPED_DATA = 0x1004,
	DPP_ATTR_I_NONCE = 0x1005,
	DPP_ATTR_I_CAPABILITIES = 0x1006,
	DPP_ATTR_R_NONCE = 0x1007,
	DPP_ATTR_R_CAPABILITIES = 0x1008,
	DPP_ATTR_R_PROTOCOL_KEY = 0x1009,
	DPP_ATTR_I_AUTH_TAG = 0x100A,
	DPP_ATTR_R_AUTH_TAG = 0x100B,
	DPP_ATTR_CONFIG_OBJ = 0x100C,
	DPP_ATTR_CONNECTOR = 0x100D,
	DPP_ATTR_CONFIG_ATTR_OBJ = 0x100E,
	DPP_ATTR_BOOTSTRAP_KEY = 0x100F,
	DPP_ATTR_OWN_NET_NK_HASH = 0x1011,
	DPP_ATTR_FINITE_CYCLIC_GROUP = 0x1012,
	DPP_ATTR_ENCRYPTED_KEY = 0x1013,
	DPP_ATTR_ENROLLEE_NONCE = 0x1014,
	DPP_ATTR_CODE_IDENTIFIER = 0x1015,
	DPP_ATTR_TRANSACTION_ID = 0x1016,
	DPP_ATTR_BOOTSTRAP_INFO = 0x1017,
	DPP_ATTR_CHANNEL = 0x1018,
	DPP_ATTR_PROTOCOL_VERSION = 0x1019,
	DPP_ATTR_ENVELOPED_DATA = 0x101A,
	DPP_ATTR_SEND_CONN_STATUS = 0x101B,
	DPP_ATTR_CONN_STATUS = 0x101C,
	DPP_ATTR_RECONFIG_FLAGS = 0x101D,
	DPP_ATTR_C_SIGN_KEY_HASH = 0x101E,
	DPP_ATTR_CSR_ATTR_REQ = 0x101F,
	DPP_ATTR_A_NONCE = 0x1020,
	DPP_ATTR_E_PRIME_ID = 0x1021,
	DPP_ATTR_CONFIGURATOR_NONCE = 0x1022,
};

enum dpp_responder_state {
	DPP_BOOSTRAPPING,
	DPP_AUTHENTICATING,
	DPP_AUTHENTICATED
};

/* taken from dpp.h */
enum dpp_bootstrap_type {
	DPP_BOOTSTRAP_QR_CODE,
	DPP_BOOTSTRAP_PKEX,
	DPP_BOOTSTRAP_NFC_URI,
};

struct dpp_frame {
	uint8_t type;
	uint16_t len;
	uint8_t *data;

	uint8_t proxy_almac[6];

	struct list_head list;
};

struct dpp_chirp {
	struct controller *c; /* TODO: lazy, should all be part of some dpp_ctx which can be traced back to cntlr */

	//uint8_t hashlen;
	uint8_t hash[32]; /* derived from DPP frame(s) */

	struct uloop_fd chirp_fd;
	struct list_head list;
	struct atimer timeout;
};

struct dpp_chan {
	uint8_t channel;
	uint8_t opclass;
	uint32_t freq;
};

/* TODO: this struct should be linked against a node? */
#define DPP_MAX_CH_OPCLASS 16
struct dpp_enrollee {
	struct controller *c; /* TODO: lazy, should all be part of some dpp_ctx which can be traced back to cntlr */

	uint8_t mac[6];
	char uri[128];

	bool onboarded;
	uint8_t bootstrap_id;

	enum wifi_band band;
	int num_chan;
	struct dpp_chan chan[DPP_MAX_CH_OPCLASS];
	enum dpp_bootstrap_type type;

	struct dpp_chirp *chirp;

	struct list_head framelist;
	struct list_head list;
};

struct dpp_context {
	struct list_head enrolleelist;
	struct list_head chirplist;
};

struct controller;

uint8_t *dpp_get_attr(uint8_t *buf, size_t len, uint16_t req_id,
		      uint16_t *ret_len);

const char *dpp_type2str(enum dpp_bootstrap_type type);
const char *dpp_frame_type2str(uint8_t type);
const char *dpp_state2str(enum dpp_responder_state state);

uint8_t dpp_get_pub_af_type(uint8_t *frame, uint16_t framelen);
uint8_t dpp_get_gas_frame_type(const uint8_t *frame, uint16_t framelen);

struct dpp_chirp *dpp_chirp_get(struct controller *c, uint8_t *hash);
struct dpp_chirp *dpp_chirp_alloc(struct controller *c, uint8_t *hash);


struct dpp_enrollee *dpp_enrollee_get(struct controller *c, uint8_t *mac);
struct dpp_enrollee *dpp_enrollee_alloc(struct controller *c, uint8_t *mac);
void dpp_enrollee_free(struct dpp_enrollee *e);

int cntlr_dpp_gen_uri(struct controller *c);
int cntlr_dpp_del_uri(struct controller *c);
int cntlr_dpp_chirp(struct controller *c);
int cntlr_dpp_listen(struct controller *c);
void dpp_sta_event_handler(void *c, struct blob_attr *msg);

int dpp_enrollee_update_hash(struct controller *c,
			     struct dpp_enrollee *e,
			     uint8_t len, uint8_t *hash);
struct dpp_frame *dpp_enrollee_get_frame(struct controller *c,
					 struct dpp_enrollee *e, uint8_t type);
struct dpp_frame *dpp_enrollee_add_frame(struct controller *c, struct dpp_enrollee *e,
			   uint8_t type, uint16_t len, uint8_t *data, uint8_t *almac);

int dpp_parse_uri_chan_list(struct controller *c, struct dpp_enrollee *e,
			    char *uri);
enum wifi_band wifi_channel_to_band(uint8_t channel);

uint16_t dpp_get_controller_tcp_port(struct controller *c, enum wifi_band band);

int dpp_append_uri_file(struct controller *c, char *uri);
struct dpp_enrollee *dpp_process_new_uri(struct controller *c, char *uri);
int dpp_cntlr_read_uris(struct controller *c);

void dpp_event_uloop_cb(struct uloop_fd *fd, unsigned int events);
#endif
